from dataclasses import dataclass, field

from markupsafe import escape

# See https://developer.mozilla.org/en-US/docs/Glossary/Void_element
VOID_ELEMENTS = frozenset(
    [
        "area",
        "base",
        "br",
        "col",
        "embed",
        "hr",
        "img",
        "input",
        "link",
        "meta",
        "param",
        "source",
        "track",
        "wbr",
    ]
)


CDATA_CONTENT_ELEMENTS = frozenset(["script", "style"])
RCDATA_CONTENT_ELEMENTS = frozenset(["textarea", "title"])
CONTENT_ELEMENTS = CDATA_CONTENT_ELEMENTS | RCDATA_CONTENT_ELEMENTS

# FUTURE: add a pretty-printer to nodes for debugging
# FUTURE: make nodes frozen (and have the parser work with mutable builders)


@dataclass(slots=True)
class Node:
    def __html__(self) -> str:
        """Return the HTML representation of the node."""
        # By default, just return the string representation
        return str(self)


@dataclass(slots=True)
class Text(Node):
    text: str

    def __str__(self) -> str:
        # Use markupsafe's escape to handle HTML escaping
        return escape(self.text)


@dataclass(slots=True)
class Fragment(Node):
    children: list[Node] = field(default_factory=list)

    def __str__(self) -> str:
        return "".join(str(child) for child in self.children)


@dataclass(slots=True)
class Comment(Node):
    text: str

    def __str__(self) -> str:
        return f"<!--{self.text}-->"


@dataclass(slots=True)
class DocumentType(Node):
    text: str = "html"

    def __str__(self) -> str:
        return f"<!DOCTYPE {self.text}>"


@dataclass(slots=True)
class Element(Node):
    tag: str
    attrs: dict[str, str | None] = field(default_factory=dict)
    children: list[Node] = field(default_factory=list)

    def __post_init__(self):
        """Ensure all preconditions are met."""
        if not self.tag:
            raise ValueError("Element tag cannot be empty.")

        # Void elements cannot have children
        if self.is_void and self.children:
            raise ValueError(f"Void element <{self.tag}> cannot have children.")

    @property
    def is_void(self) -> bool:
        return self.tag in VOID_ELEMENTS

    @property
    def is_content(self) -> bool:
        return self.tag in CONTENT_ELEMENTS

    def __str__(self) -> str:
        # We use markupsafe's escape to handle HTML escaping of attribute values
        # which means it's possible to mark them as safe if needed.
        attrs_str = "".join(
            f" {key}" if value is None else f' {key}="{escape(value)}"'
            for key, value in self.attrs.items()
        )
        if self.is_void:
            return f"<{self.tag}{attrs_str} />"
        if not self.children:
            return f"<{self.tag}{attrs_str}></{self.tag}>"
        if self.is_content:
            # Content elements should *not* escape their content when
            # rendering to HTML. Sheesh, HTML is weird.
            children_str = "".join(
                child.text if isinstance(child, Text) else str(child)
                for child in self.children
            )
        else:
            children_str = "".join(str(child) for child in self.children)
        return f"<{self.tag}{attrs_str}>{children_str}</{self.tag}>"



import random
import string
import typing as t
from collections.abc import Iterable
from functools import lru_cache
from string.templatelib import Interpolation, Template

from markupsafe import Markup

from .classnames import classnames
from .nodes import Element, Fragment, Node, Text
from .parser import parse_html
from .utils import format_interpolation as base_format_interpolation


@t.runtime_checkable
class HasHTMLDunder(t.Protocol):
    def __html__(self) -> str: ...


# --------------------------------------------------------------------------
# Value formatting
# --------------------------------------------------------------------------


def _format_safe(value: object, format_spec: str) -> str:
    """Use Markup() to mark a value as safe HTML."""
    assert format_spec == "safe"
    return Markup(value)


def _format_unsafe(value: object, format_spec: str) -> str:
    """Convert a value to a plain string, forcing it to be treated as unsafe."""
    assert format_spec == "unsafe"
    return str(value)


CUSTOM_FORMATTERS = (("safe", _format_safe), ("unsafe", _format_unsafe))


def format_interpolation(interpolation: Interpolation) -> object:
    return base_format_interpolation(
        interpolation,
        formatters=CUSTOM_FORMATTERS,
    )


# --------------------------------------------------------------------------
# Instrumentation, Parsing, and Caching
# --------------------------------------------------------------------------

_PLACEHOLDER_PREFIX = f"t🐍-{''.join(random.choices(string.ascii_lowercase, k=4))}-"
_PP_LEN = len(_PLACEHOLDER_PREFIX)


def _placeholder(i: int) -> str:
    """Generate a placeholder for the i-th interpolation."""
    return f"{_PLACEHOLDER_PREFIX}{i}"


def _placholder_index(s: str) -> int:
    """Extract the index from a placeholder string."""
    return int(s[_PP_LEN:])


def _instrument(
    strings: tuple[str, ...], callable_ids: tuple[int | None, ...]
) -> t.Iterable[str]:
    """
    Join the strings with placeholders in between where interpolations go.

    This is used to prepare the template string for parsing, so that we can
    later substitute the actual interpolated values into the parse tree.

    The placeholders are chosen to be unlikely to collide with typical HTML
    content.
    """
    count = len(strings)

    callable_placeholders: dict[int, str] = {}

    for i, s in enumerate(strings):
        yield s
        # There are always count-1 placeholders between count strings.
        if i < count - 1:
            placeholder = _placeholder(i)

            # Special case for component callables: if the interpolation
            # is a callable, we need to make sure that any matching closing
            # tag uses the same placeholder.
            callable_id = callable_ids[i]
            if callable_id:
                placeholder = callable_placeholders.setdefault(callable_id, placeholder)

            yield placeholder


@lru_cache()
def _instrument_and_parse_internal(
    strings: tuple[str, ...], callable_ids: tuple[int | None, ...]
) -> Node:
    """
    Instrument the strings and parse the resulting HTML.

    The result is cached to avoid re-parsing the same template multiple times.
    """
    instrumented = _instrument(strings, callable_ids)
    return parse_html(instrumented)


def _callable_id(value: object) -> int | None:
    """Return a unique identifier for a callable, or None if not callable."""
    return id(value) if callable(value) else None


def _instrument_and_parse(template: Template) -> Node:
    """Instrument and parse a template, returning a tree of Nodes."""
    # This is a thin wrapper around the cached internal function that does the
    # actual work. This exists to handle the syntax we've settled on for
    # component invocation, namely that callables are directly included as
    # interpolations both in the open *and* the close tags. We need to make
    # sure that matching tags... match!
    #
    # If we used `tdom`'s approach of component closing tags of <//> then we
    # wouldn't have to do this. But I worry that tdom's syntax is harder to read
    # (it's easy to miss the closing tag) and may prove unfamiliar for
    # users coming from other templating systems.
    callable_ids = tuple(
        _callable_id(interpolation.value) for interpolation in template.interpolations
    )
    return _instrument_and_parse_internal(template.strings, callable_ids)


# --------------------------------------------------------------------------
# Placeholder Substitution
# --------------------------------------------------------------------------


def _force_dict(value: t.Any, *, kind: str) -> dict:
    """Try to convert a value to a dict, raising TypeError if not possible."""
    try:
        return dict(value)
    except (TypeError, ValueError):
        raise TypeError(
            f"Cannot use {type(value).__name__} as value for {kind} attributes"
        ) from None


def _substitute_aria_attrs(value: object) -> t.Iterable[tuple[str, str | None]]:
    """Produce aria-* attributes based on the interpolated value for "aria"."""
    d = _force_dict(value, kind="aria")
    for sub_k, sub_v in d.items():
        if sub_v is True:
            yield f"aria-{sub_k}", "true"
        elif sub_v is False:
            yield f"aria-{sub_k}", "false"
        elif sub_v is None:
            pass
        else:
            yield f"aria-{sub_k}", str(sub_v)


def _substitute_data_attrs(value: object) -> t.Iterable[tuple[str, str | None]]:
    """Produce data-* attributes based on the interpolated value for "data"."""
    d = _force_dict(value, kind="data")
    for sub_k, sub_v in d.items():
        if sub_v is True:
            yield f"data-{sub_k}", None
        elif sub_v not in (False, None):
            yield f"data-{sub_k}", str(sub_v)


def _substitute_class_attr(value: object) -> t.Iterable[tuple[str, str | None]]:
    """Substitute a class attribute based on the interpolated value."""
    yield ("class", classnames(value))


def _substitute_style_attr(value: object) -> t.Iterable[tuple[str, str | None]]:
    """Substitute a style attribute based on the interpolated value."""
    try:
        d = _force_dict(value, kind="style")
        style_str = "; ".join(f"{k}: {v}" for k, v in d.items())
        yield ("style", style_str)
    except TypeError:
        yield ("style", str(value))


def _substitute_spread_attrs(
    value: object,
) -> t.Iterable[tuple[str, object | None]]:
    """
    Substitute a spread attribute based on the interpolated value.

    A spread attribute is one where the key is a placeholder, indicating that
    the entire attribute set should be replaced by the interpolated value.
    The value must be a dict or iterable of key-value pairs.
    """
    d = _force_dict(value, kind="spread")
    for sub_k, sub_v in d.items():
        yield from _substitute_attr(sub_k, sub_v)


# A collection of custom handlers for certain attribute names that have
# special semantics. This is in addition to the special-casing in
# _substitute_attr() itself.
CUSTOM_ATTR_HANDLERS = {
    "class": _substitute_class_attr,
    "data": _substitute_data_attrs,
    "style": _substitute_style_attr,
    "aria": _substitute_aria_attrs,
}


def _substitute_attr(
    key: str,
    value: object,
) -> t.Iterable[tuple[str, object | None]]:
    """
    Substitute a single attribute based on its key and the interpolated value.

    A single parsed attribute with a placeholder may result in multiple
    attributes in the final output, for instance if the value is a dict or
    iterable of key-value pairs. Likewise, a value of False will result in
    the attribute being omitted entirely; nothing is yielded in that case.
    """
    # Special handling for certain attribute names that have special semantics
    if custom_handler := CUSTOM_ATTR_HANDLERS.get(key):
        yield from custom_handler(value)
        return

    # General handling for all other attributes:
    match value:
        case True:
            yield (key, None)
        case False | None:
            pass
        case _:
            yield (key, value)


def _substitute_attrs(
    attrs: dict[str, str | None], interpolations: tuple[Interpolation, ...]
) -> dict[str, object | None]:
    """Substitute placeholders in attributes based on the corresponding interpolations."""
    new_attrs: dict[str, object | None] = {}
    for key, value in attrs.items():
        if value and value.startswith(_PLACEHOLDER_PREFIX):
            index = _placholder_index(value)
            interpolation = interpolations[index]
            value = format_interpolation(interpolation)
            for sub_k, sub_v in _substitute_attr(key, value):
                new_attrs[sub_k] = sub_v
        elif key.startswith(_PLACEHOLDER_PREFIX):
            index = _placholder_index(key)
            interpolation = interpolations[index]
            value = format_interpolation(interpolation)
            for sub_k, sub_v in _substitute_spread_attrs(value):
                new_attrs[sub_k] = sub_v
        else:
            new_attrs[key] = value
    return new_attrs


def _substitute_and_flatten_children(
    children: t.Iterable[Node], interpolations: tuple[Interpolation, ...]
) -> list[Node]:
    """Substitute placeholders in a list of children and flatten any fragments."""
    new_children: list[Node] = []
    for child in children:
        substituted = _substitute_node(child, interpolations)
        if isinstance(substituted, Fragment):
            # This can happen if an interpolation results in a Fragment, for
            # instance if it is iterable.
            new_children.extend(substituted.children)
        else:
            new_children.append(substituted)
    return new_children


def _node_from_value(value: object) -> Node:
    """
    Convert an arbitrary value to a Node.

    This is the primary substitution performed when replacing interpolations
    in child content positions.
    """
    match value:
        case str():
            return Text(value)
        case Node():
            return value
        case Template():
            return html(value)
        case False:
            return Text("")
        case Iterable():
            children = [_node_from_value(v) for v in value]
            return Fragment(children=children)
        case HasHTMLDunder():
            # CONSIDER: could we return a lazy Text?
            return Text(Markup(value.__html__()))
        case _:
            # CONSIDER: could we return a lazy Text?
            return Text(str(value))


type ComponentReturn = Node | Template | str | HasHTMLDunder
type ComponentCallable = t.Callable[..., ComponentReturn | t.Iterable[ComponentReturn]]


def _invoke_component(
    tag: str,
    new_attrs: dict[str, object | None],
    new_children: list[Node],
    interpolations: tuple[Interpolation, ...],
) -> Node:
    """Substitute a component invocation based on the corresponding interpolations."""
    index = _placholder_index(tag)
    interpolation = interpolations[index]
    value = format_interpolation(interpolation)
    # TODO: consider use of signature() or other approaches to validation.
    if not callable(value):
        raise TypeError(
            f"Expected a callable for component invocation, got {type(value).__name__}"
        )
    # Replace attr names hyphens with underscores for Python kwargs
    kwargs = {k.replace("-", "_"): v for k, v in new_attrs.items()}

    # Call the component and return the resulting node
    result = value(*new_children, **kwargs)
    match result:
        case Node():
            return result
        case Template():
            return html(result)
        case str():
            return Text(result)
        case HasHTMLDunder():
            return Text(Markup(result.__html__()))
        case _:
            raise TypeError(
                f"Component callable must return a Node, Template, or str; "
                f"got {type(result).__name__}"
            )


def _stringify_attrs(attrs: dict[str, object | None]) -> dict[str, str | None]:
    """Convert all attribute values to strings, preserving None values."""
    return {k: str(v) if v is not None else None for k, v in attrs.items()}


def _substitute_node(p_node: Node, interpolations: tuple[Interpolation, ...]) -> Node:
    """Substitute placeholders in a node based on the corresponding interpolations."""
    match p_node:
        case Text(text) if str(text).startswith(_PLACEHOLDER_PREFIX):
            index = _placholder_index(str(text))
            interpolation = interpolations[index]
            value = format_interpolation(interpolation)
            return _node_from_value(value)
        case Element(tag=tag, attrs=attrs, children=children):
            new_attrs = _substitute_attrs(attrs, interpolations)
            new_children = _substitute_and_flatten_children(children, interpolations)
            if tag.startswith(_PLACEHOLDER_PREFIX):
                return _invoke_component(tag, new_attrs, new_children, interpolations)
            else:
                new_attrs = _stringify_attrs(new_attrs)
                return Element(tag=tag, attrs=new_attrs, children=new_children)
        case Fragment(children=children):
            new_children = _substitute_and_flatten_children(children, interpolations)
            return Fragment(children=new_children)
        case _:
            return p_node


# --------------------------------------------------------------------------
# Public API
# --------------------------------------------------------------------------


def html(template: Template) -> Node:
    """Parse a t-string and return a tree of Nodes."""
    # Parse the HTML, returning a tree of nodes with placeholders
    # where interpolations go.
    p_node = _instrument_and_parse(template)
    return _substitute_node(p_node, template.interpolations)



import typing as t
from html.parser import HTMLParser

from .nodes import VOID_ELEMENTS, Comment, DocumentType, Element, Fragment, Node, Text


class NodeParser(HTMLParser):
    root: Fragment
    stack: list[Element]

    def __init__(self):
        super().__init__()
        self.root = Fragment(children=[])
        self.stack = []

    def handle_starttag(
        self, tag: str, attrs: t.Sequence[tuple[str, str | None]]
    ) -> None:
        element = Element(tag, attrs=dict(attrs), children=[])
        self.stack.append(element)

        # Unfortunately, Python's built-in HTMLParser has inconsistent behavior
        # with void elements. In particular, it calls handle_endtag() for them
        # only if they explicitly self-close (e.g., <br />). But in the HTML
        # spec itself, *there is no distinction* between <br> and <br />.
        # So we need to handle this case ourselves.
        #
        # See https://github.com/python/cpython/issues/69445
        if tag in VOID_ELEMENTS:
            # Always call handle_endtag for void elements. If it happens
            # to be self-closed in the input, handle_endtag() will effectively
            # be called twice. We ignore the second call there.
            self.handle_endtag(tag)

    def handle_endtag(self, tag: str) -> None:
        if tag in VOID_ELEMENTS:
            # Special case: handle Python issue #69445 (see comment above).
            most_recent_closed = self.get_most_recent_closed_element()
            if most_recent_closed and most_recent_closed.tag == tag:
                # Ignore this call; we've already closed it.
                return
            open_element = self.get_open_element()
            if open_element and open_element.tag == tag:
                _ = self.stack.pop()
                self.append_child(open_element)
                return

        if not self.stack:
            raise ValueError(f"Unexpected closing tag </{tag}> with no open element.")

        element = self.stack.pop()
        if element.tag != tag:
            raise ValueError(f"Mismatched closing tag </{tag}> for <{element.tag}>.")

        self.append_child(element)

    def handle_data(self, data: str) -> None:
        text = Text(data)
        self.append_child(text)

    def handle_comment(self, data: str) -> None:
        comment = Comment(data)
        self.append_child(comment)

    def handle_decl(self, decl: str) -> None:
        if decl.upper().startswith("DOCTYPE"):
            doctype_content = decl[7:].strip()
            doctype = DocumentType(doctype_content)
            self.append_child(doctype)
        # For simplicity, we ignore other declarations.
        pass

    def get_parent(self) -> Fragment | Element:
        """Return the current parent node to which new children should be added."""
        return self.stack[-1] if self.stack else self.root

    def get_open_element(self) -> Element | None:
        """Return the currently open Element, if any."""
        return self.stack[-1] if self.stack else None

    def get_most_recent_closed_element(self) -> Element | None:
        """Return the most recently closed Element, if any."""
        parent = self.get_parent()
        if parent.children and isinstance(parent.children[-1], Element):
            return parent.children[-1]
        return None

    def append_child(self, child: Node) -> None:
        parent = self.get_parent()
        # We *know* our parser is using lists for children, so this cast is safe.
        parent.children.append(child)

    def close(self) -> None:
        if self.stack:
            raise ValueError("Invalid HTML structure: unclosed tags remain.")
        super().close()

    def get_node(self) -> Node:
        """Get the Node tree parsed from the input HTML."""
        # CONSIDER: Should we invert things and offer streaming parsing?
        assert not self.stack, "Did you forget to call close()?"
        if len(self.root.children) > 1:
            # The parse structure results in multiple root elements, so we
            # return a Fragment to hold them all.
            return self.root
        elif len(self.root.children) == 1:
            # The parse structure results in a single root element, so we
            # return that element directly. This will be a non-Fragment Node.
            return self.root.children[0]
        else:
            # Special case: the parse structure is empty; we treat
            # this as an empty Text Node.
            return Text("")


def parse_html(input: str | t.Iterable[str]) -> Node:
    """
    Parse a string, or sequence of HTML string chunks, into a Node tree.

    If a single string is provided, it is parsed as a whole. If an iterable
    of strings is provided, each string is fed to the parser in sequence.
    This is particularly useful if you want to keep specific text chunks
    separate in the resulting Node tree.
    """
    parser = NodeParser()
    iterable = [input] if isinstance(input, str) else input
    for chunk in iterable:
        parser.feed(chunk)
    parser.close()
    return parser.get_node()



import typing as t
from string.templatelib import Interpolation


@t.overload
def convert[T](value: T, conversion: None) -> T: ...


@t.overload
def convert(value: object, conversion: t.Literal["a", "r", "s"]) -> str: ...


def convert[T](value: T, conversion: t.Literal["a", "r", "s"] | None) -> T | str:
    """
    Convert a value according to the given conversion specifier.

    In the future, something like this should probably ship with Python itself.
    """
    if conversion == "a":
        return ascii(value)
    elif conversion == "r":
        return repr(value)
    elif conversion == "s":
        return str(value)
    else:
        return value


type FormatMatcher = t.Callable[[str], bool]
"""A predicate function that returns True if a given format specifier matches its criteria."""

type CustomFormatter = t.Callable[[object, str], str]
"""A function that takes a value and a format specifier and returns a formatted string."""

type MatcherAndFormatter = tuple[str | FormatMatcher, CustomFormatter]
"""
A pair of a matcher and its corresponding formatter.

The matcher is used to determine if the formatter should be applied to a given 
format specifier. If the matcher is a string, it must exactly match the format
specifier. If it is a FormatMatcher, it is called with the format specifier and
should return True if the formatter should be used.
"""


def _matcher_matches(matcher: str | FormatMatcher, format_spec: str) -> bool:
    """Check if a matcher matches a given format specifier."""
    return matcher == format_spec if isinstance(matcher, str) else matcher(format_spec)


def _format_interpolation(
    value: object,
    format_spec: str,
    conversion: t.Literal["a", "r", "s"] | None,
    *,
    formatters: t.Sequence[MatcherAndFormatter],
) -> object:
    converted = convert(value, conversion)
    if format_spec:
        for matcher, formatter in formatters:
            if _matcher_matches(matcher, format_spec):
                return formatter(converted, format_spec)
        return format(converted, format_spec)
    return converted


def format_interpolation(
    interpolation: Interpolation,
    *,
    formatters: t.Sequence[MatcherAndFormatter] = tuple(),
) -> object:
    """
    Format an Interpolation's value according to its format spec and conversion.

    PEP 750 allows t-string processing code to decide whether, and how, to
    interpret format specifiers. This function takes an optional sequence of
    (matcher, formatter) pairs. If a matcher returns True for the given format
    spec, the corresponding formatter is used to format the value. If no
    matchers match, the default formatting behavior is used.

    Conversions are always applied before formatting.
    """
    return _format_interpolation(
        interpolation.value,
        interpolation.format_spec,
        interpolation.conversion,
        formatters=formatters,
    )



def classnames(*args: object) -> str:
    """
    Construct a space-separated class string from various inputs.

    Accepts strings, lists/tuples of strings, and dicts mapping class names to
    boolean values. Ignores None and False values.

    Examples:
        classnames("btn", "btn-primary") -> "btn btn-primary"
        classnames("btn", {"btn-primary": True, "disabled": False}) -> "btn btn-primary"
        classnames(["btn", "btn-primary"], {"disabled": True}) -> "btn btn-primary disabled"
        classnames("btn", None, False, "active") -> "btn active"

    Args:
        *args: Variable length argument list containing strings, lists/tuples,
               or dicts.

    Returns:
        A single string with class names separated by spaces.
    """
    classes: list[str] = []
    # Use a queue to process arguments iteratively, preserving order.
    queue = list(args)

    while queue:
        arg = queue.pop(0)

        if not arg:  # Handles None, False, empty strings/lists/dicts
            continue

        if isinstance(arg, str):
            classes.append(arg)
        elif isinstance(arg, dict):
            for key, value in arg.items():
                if value:
                    classes.append(key)
        elif isinstance(arg, (list, tuple)):
            # Add items to the front of the queue to process them next, in order.
            queue[0:0] = arg
        elif isinstance(arg, bool):
            pass  # Explicitly ignore booleans not in a dict
        else:
            raise ValueError(f"Invalid class argument type: {type(arg).__name__}")

    # Filter out empty strings and join the result.
    return " ".join(stripped for c in classes if (stripped := c.strip()))



from markupsafe import Markup, escape

from .nodes import Comment, DocumentType, Element, Fragment, Text
from .processor import ComponentCallable, html

# We consider `Markup` and `escape` to be part of this module's public API

__all__ = [
    "Comment",
    "ComponentCallable",
    "DocumentType",
    "Element",
    "escape",
    "Fragment",
    "html",
    "Markup",
    "Text",
]
